package com.bc.beam.visat;

import com.bc.progress.ProgressController;
import com.bc.progress.ProgressControllerPool;
import com.bc.swing.ProgressMonitor;
import org.esa.beam.framework.datamodel.Band;
import org.esa.beam.framework.datamodel.GeoCoding;
import org.esa.beam.framework.datamodel.Pin;
import org.esa.beam.framework.datamodel.Product;
import org.esa.beam.framework.datamodel.RasterDataNode;
import org.esa.beam.framework.ui.ModalDialog;
import org.esa.beam.framework.ui.SelectExportMethodDialog;
import org.esa.beam.framework.ui.SwingWorker;
import org.esa.beam.framework.ui.UIUtils;
import org.esa.beam.framework.ui.command.Command;
import org.esa.beam.framework.ui.command.CommandAdapter;
import org.esa.beam.framework.ui.command.CommandEvent;
import org.esa.beam.framework.ui.command.CommandListener;
import org.esa.beam.framework.ui.command.CommandManager;
import org.esa.beam.framework.ui.product.ProductSceneView;
import org.esa.beam.util.SystemUtils;
import org.esa.beam.util.io.FileUtils;
import org.esa.beam.visat.VisatApp;
import org.esa.beam.visat.VisatPlugIn;
import org.esa.beam.visat.modules.exprt.ExportROIPixelsVPI;

import javax.swing.JOptionPane;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;

/*
 * Copyright (C) 2005  by Maximilian Aulinger
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation;
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/**
 * Installs a plug-in wich can be used to export all pixel values within a region around a pin.
 *
 * @author Maximilian Aulinger
 */
public class ExportPinPixels implements VisatPlugIn {

    private final static String COMMAND_ID = "exportPinPixels";
    private final static String COMMAND_NAME = "Export Pin Pixels";

    private VisatApp _visatApp;
    public static final String ERR_MSG_BASE = "Pin pixels cannot be exported:\n";

    public void initPlugIn(final VisatApp visatApp) {
        _visatApp = visatApp;

// Get VISAT's command manager

        final CommandManager commandManager = visatApp.getCommandManager();

// Create a new VISAT command handler (as instance of an anonymous class)
// that lets user export the relevant pixels.

        final CommandListener commandListener = new CommandAdapter() {

            /**
             * Invoked when a command action is performed.
             *
             * @param event the command event
             */
            public void actionPerformed(final CommandEvent event) {
                exportPinPixels(event);

            }

            /**
             * Called when a command should update its state.
             * <p/>
             * <p> This method can contain some code which analyzes the underlying element and makes a decision whether
             * this item or group should be made visible/invisible or enabled/disabled etc.
             *
             * @param event the command event
             */
            public void updateState(final CommandEvent event) {
                final ProductSceneView view = visatApp.getSelectedProductSceneView();
                final boolean enabled = view != null;
//                boolean enabled = isROIDefinedInSelectedRaster();
                event.getCommand().setEnabled(enabled);
            }
        };

// Now create the actual command with the command handler we've just created above.
// We use some of the multiple command property setters here to configure our command.

        final Command command = commandManager.createExecCommand(COMMAND_ID, commandListener);
        command.setEnabled(false); // disable command
        command.setText(COMMAND_NAME + "..."); // e.g. menu item text
        command.setShortDescription(COMMAND_NAME); // e.g. text used as tooltip text
        command.setLongDescription(COMMAND_NAME); // e.g. text shown in status bar
        command.setParent("export"); // Place in "export" command group
        command.setContexts(new String[]{"image"}); // Show command, if context is "image" (context menu)

    }


    public void updateComponentTreeUI() {
// No code required here

    }

    /**
     * Performs the "Export Pin Pixels" command.
     */
    private void exportPinPixels(final CommandEvent event) {


        // Get current VISAT view showing a product's band
        final ProductSceneView view = _visatApp.getSelectedProductSceneView();
        if (view == null) {
            return;
        }
        //Get the displayed raster data node (band or tie-point grid)
        final RasterDataNode raster = view.getRaster();
        //Get the selected product
        final Product product = raster.getProduct();
        final Pin myPin = product.getSelectedPin();

        // The "Export Pin Pixels" command is canceled if there is no pin defined or selected in the given product
        if (myPin == null) {
            _visatApp.showErrorDialog(COMMAND_NAME, ERR_MSG_BASE + "There is no pin defined or selected.");
            return;
        }

        // The "Export Pin Pixels" command is canceled if the possible region size is smaller than 3 x 3 pixels
        if (!selectedPinCanBeExported(product, myPin)) {
            _visatApp.showErrorDialog(COMMAND_NAME,
                                      ERR_MSG_BASE + "The position of the selected pin is to close to the product's border.");
            return;
        }

        final ExportPinPixelsDialog dialog = new ExportPinPixelsDialog(_visatApp, product, myPin);
        // shows a dialog which lets the user specify the region he wants to export.
        final int dialogAnswer = dialog.show();

        // return if user pressed the Cancel-button
        if (dialogAnswer != ModalDialog.ID_OK) {
            return;
        }

        // gets the com.bc.beam.visat.RegionAroundPin the user has defined in the dialog window
        final RegionAroundPin exportRegion = dialog.getDefinedRegion();

        final int numPixels = exportRegion.getNumberOfRelevantPixels();

        final String questionText = "How do you want to export the pixel values?\n";

        final String numPixelsText;
        if (numPixels == 1) {
            numPixelsText = "One pixel will be exported.\n";
        } else {
            numPixelsText = numPixels + " pixels will be exported.\n";
        }


// Get export method from user
        final int method = SelectExportMethodDialog.run(_visatApp.getMainFrame(), getWindowTitle(),
                                                        questionText + numPixelsText, event.getCommand().getHelpId());

        final PrintWriter out;
        final StringBuffer clipboardText;
        final int initialBufferSize = 256000;
        if (method == SelectExportMethodDialog.EXPORT_TO_CLIPBOARD) {
// Write into string buffer
            final StringWriter stringWriter = new StringWriter(initialBufferSize);
            out = new PrintWriter(stringWriter);
            clipboardText = stringWriter.getBuffer();
        } else if (method == SelectExportMethodDialog.EXPORT_TO_FILE) {
// Write into file, get file from user
            final File file = promptForFile(_visatApp, createDefaultFileName(product));
            if (file == null) {
                return; // Cancel
            }
            final FileWriter fileWriter;
            try {
                fileWriter = new FileWriter(file);
            } catch (IOException e) {
                _visatApp.showErrorDialog(COMMAND_NAME,
                                          ERR_MSG_BASE + "Failed to create file '" + file + "':\n" + e.getMessage());
                return; // Error
            }
            out = new PrintWriter(new BufferedWriter(fileWriter, initialBufferSize));
            clipboardText = null;
        } else {
            return; // Cancel
        }
// Create a progress monitor and adds them to the progress controller pool in order to show export progress
        final ProgressMonitor monitor = new ProgressMonitor(_visatApp.getMainFrame(), COMMAND_NAME);
// Create a new swing worker instance (as instance of an anonymous class).
// When the swing worker's start() method is called, a new separate thread is started.
// The swing worker's construct() method is executed in that thread, so that
// VISAT can keep on handling user events.
//
        final SwingWorker swingWorker = new SwingWorker(monitor) {

            /**
             * Compute the value to be returned by the <code>get</code> method. This method is executed in a separate
             * worker thread.
             */
            public Object construct() {
                try {
                    if (exportPinPixels(product, out, exportRegion) && clipboardText != null) {
                        SystemUtils.copyToClipboard(clipboardText.toString());
                        clipboardText.setLength(0);
                    }
                } catch (Exception e) {
                    return e;
                } finally {
                    out.close();
                }
                return null;
            }

            /**
             * Called on the event dispatching thread (not on the worker thread) after the <code>construct</code> method
             * has returned.
             */
            public void finished() {
// clear status bar
                _visatApp.clearStatusBarMessage();
// show default-cursor
                UIUtils.setRootFrameDefaultCursor(_visatApp.getMainFrame());
// On error, show error message
                if (getValue() instanceof Exception) {
                    final Exception e = (Exception) getValue();
                    _visatApp.showErrorDialog(COMMAND_NAME,
                                              ERR_MSG_BASE + e.getMessage());
                }
            }
        };

// show wait-cursor
        UIUtils.setRootFrameWaitCursor(_visatApp.getMainFrame());
// show message in status bar
        _visatApp.setStatusBarMessage("Exporting Pin pixels...");

// Start separate worker thread.
        swingWorker.start();
    }

    /**
     * Opens a modal file chooser dialog that prompts the user to select the output file name.
     *
     * @param visatApp the VISAT application
     *
     * @return the selected file, <code>null</code> means "Cancel"
     */
    private File promptForFile(final VisatApp visatApp, final String defaultFileName) {
        // Loop while the user does not want to overwrite a selected, existing file
        // or if the user presses "Cancel"

        File file = null;
        while (file == null) {
            file = visatApp.showFileSaveDialog(COMMAND_NAME,
                                               false,
                                               null,
                                               ".txt",
                                               defaultFileName,
                                               COMMAND_ID + ".lastDir");
            if (file == null) {
                return null; // Cancel
            } else if (file.exists()) {
                final int status = JOptionPane.showConfirmDialog(visatApp.getMainFrame(),
                                                           "The file '" + file + "' already exists.\n" +
                                                           "Overwrite it?",
                                                           "VISAT - " + COMMAND_NAME,
                                                           JOptionPane.YES_NO_CANCEL_OPTION,
                                                           JOptionPane.WARNING_MESSAGE);
                if (status == JOptionPane.CANCEL_OPTION) {
                    return null; // Cancel
                } else if (status == JOptionPane.NO_OPTION) {
                    file = null; // No, do not overwrite, let user select other file
                }
            }
        }
        return file;
    }

    /**
     * Creates the default filename for the exported data.
     */
    private String createDefaultFileName(final Product product) {
        return FileUtils.getFilenameWithoutExtension(product.getName()) + "_Pin.txt";
    }

    /**
     * Writes all relevant pixel values within the defined region to the specified output writer.
     *
     * @param product the product providing the pixel values
     * @param out     the data output writer
     * @param region  the userdefined regionAroundPin
     *
     * @return	 <code>true</code> for success, <code>false</code> if export has been terminated (by user)
     */
    private boolean exportPinPixels(final Product product, final PrintWriter out, final RegionAroundPin region) throws IOException {
        // get all bands of the given product
        final Band[] bands = product.getBands();
        final GeoCoding geoCoding = product.getGeoCoding();

        // get the coordinates of the first pixel
        final int firstX = (int) region.getFirstPixel().getX();
        final int firstY = (int) region.getFirstPixel().getY();

        // get the coordinates of the last pixel
        final int lastX = (int) region.getLastPixel().getX();
        final int lastY = (int) region.getLastPixel().getY();


        writePinPixelsHeaderLine(out, region);

        //	write the headline and bandnames into the data output writer
        ExportROIPixelsVPI.writeHeaderLine(out, geoCoding, bands);

        final ProgressController pc = ProgressControllerPool.getInstance().getProgressController();
        pc.fireProcessStarted("Writing pixel data...", 0, lastY);

        try {
            for (int y = firstY; y <= lastY; y++) {
                if (pc.isTerminationRequested()) {
                    return false;
                }
                for (int x = firstX; x <= lastX; x++) {

                    // write pixel values into the data output writer if not filtered by an arithmetic expression
                    if (region.isRelevantPixel(x, y)) {
                        ExportROIPixelsVPI.writeDataLine(out, geoCoding, bands, x, y);
                    }
                }
                pc.fireProcessInProgress(y);
            }
        } finally {
            pc.fireProcessEnded();
        }
        return true;
    }

    private String getWindowTitle() {
        return _visatApp.getAppName() + " - " + COMMAND_NAME;
    }


//    /**
//     * @param product the product providing the pixel values
//     *
//     * @return	the pin the user has selected in the given product
//     */
//    private Pin getSelectedPin(final Product product) {
//        final Pin[] allPins = product.getPins();
//        Pin selectedPin = null;
//        for (int i = 0; i < allPins.length; i++) {
//            if (allPins[i].isSelected()) {
//                selectedPin = allPins[i];
//            }
//        }
//
//        return selectedPin;
//    }

    /**
     * Writes the size of the exported region and the used expression to the output writer.
     *
     * @param out          the data output writer
     * @param exportRegion the com.bc.beam.visat.RegionAroundPin
     */
    private void writePinPixelsHeaderLine(final PrintWriter out, final RegionAroundPin exportRegion) {
        out.print("Exported region size in pixels:");
        out.print("\t");
        out.print(exportRegion.getRegion());
        out.print("\t");
        out.print("Used expression:");
        out.print("\t");
        final String usedExpression = exportRegion.getExpression();
        if (usedExpression != null) {
            out.print(usedExpression);
        } else {
            out.print("no expression used");
        }
        out.print("\t");

        out.print("\n");
        out.print("\n");

    }

    /**
     * Checks if the position of the selected pin is to close to the product's border.
     *
     * @param product     the product providing the pixel values
     * @param selectedPin the pin defining the center of the region
     *
     * @return <code>fales</code> if the pin is closer than 3 pixels to the product's border
     */
    private boolean selectedPinCanBeExported(final Product product, final Pin selectedPin) {
        final int height = product.getSceneRasterHeight();
        final int width = product.getSceneRasterWidth();
        final int pinX = (int) selectedPin.getPixelPos().getX();
        final int pinY = (int) selectedPin.getPixelPos().getY();
        if (pinX <= 2 || pinY <= 2 || pinX >= width - 2 || pinY >= height - 2) {
            return false;
        }
        return true;
    }
}