package com.bc.beam.visat;

import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;

import javax.swing.JButton;
import javax.swing.JPanel;

import org.esa.beam.framework.datamodel.Pin;
import org.esa.beam.framework.datamodel.Product;
import org.esa.beam.framework.param.ParamChangeEvent;
import org.esa.beam.framework.param.ParamChangeListener;
import org.esa.beam.framework.param.Parameter;
import org.esa.beam.framework.ui.GridBagUtils;
import org.esa.beam.framework.ui.ModalDialog;
import org.esa.beam.framework.ui.product.ProductExpressionPane;
import org.esa.beam.util.Debug;
import org.esa.beam.visat.VisatApp;

import com.bc.jexp.ParseException;
import com.bc.jexp.Parser;

/*
 * Copyright (C) 2005  by Maximilian Aulinger
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/**
 * This class defines a dialog which lets the user specify a quadratic region around a selected pin
 * and an arithmetic expression.
 * 
 * @author Maximilian Aulinger
 *
 */

public class ExportPinPixelsDialog extends ModalDialog {
    private static final String _PARAM_NAME_USE_ARTIHMETIC_EXPRESSION = "useArithmeticExpr";
	
    private final VisatApp _visatApp;
	private final Product _product;
	private final Pin _selectedPin;
	private RegionAroundPin exportRegion;
	private Parameter _paramRegion;
	private Parameter _paramExpression;
	private Parameter _paramUseExpression;
    private JButton _editExpressionButton;


	/**
	 * Initialises a new instance of this class.
	 * @param visatApp 	a reference to the VISAT application instance
	 * @param product 	the product providing the pixel values
	 * @param pin		the pin defining the center of the region
	 */
	public ExportPinPixelsDialog(final VisatApp visatApp,final Product product,final Pin pin) {
		super(visatApp.getMainFrame(), "Export Pin Pixels", ModalDialog.ID_OK_CANCEL, null);
		_product = product;
		_visatApp = visatApp;
		_selectedPin = pin;
		exportRegion = null;
		initParameter();
		createUI();
	}
	
	/**
	 * Get's the region specified by the user
	 * @return the defined region. <code>null</code> when called before user has pressed the OK-button.
	 */
	public RegionAroundPin getDefinedRegion() {
		return exportRegion;
	}
	
	/**
	 * Initialises the selectable parameters.
	 */
	private void initParameter() {
        final ParamChangeListener paramChangeListener = createParamChangeListener();
		
		final int nMax = computeMaxRegionSize(_selectedPin,_product.getSceneRasterHeight(),_product.getSceneRasterWidth());
		final String[] possibleRegions = defineRegionSelection(nMax);
						
		_paramRegion = new Parameter("exportRegion",possibleRegions[0]);
		_paramRegion.getProperties().setLabel("Select the size of the region you want to export:");
		_paramRegion.getProperties().setValueSet(possibleRegions);
        _paramRegion.getProperties().setValueSetBound(true);
				
        _paramExpression = new Parameter("arithmeticExpr", "");
        _paramExpression.getProperties().setLabel("Expression"); /*I18N*/
        _paramExpression.getProperties().setDescription("Arithmetic expression"); /*I18N*/
        _paramExpression.getProperties().setNumRows(3);
        
        _paramUseExpression = new Parameter(_PARAM_NAME_USE_ARTIHMETIC_EXPRESSION, new Boolean(true));
        _paramUseExpression.getProperties().setLabel("Use Expression"); /*I18N*/
        _paramUseExpression.addParamChangeListener(paramChangeListener);

    }
	
	 /**
	 * Creates the ui elements.
	 */
	private void createUI() {
       
        _editExpressionButton = new JButton("Edit Expression...");
        _editExpressionButton.addActionListener(createEditExpressionButtonListener());

        final JPanel panel = GridBagUtils.createPanel();
        int line = 0;
        final GridBagConstraints gbc = new GridBagConstraints();
        
        gbc.gridy = ++line;
        GridBagUtils.addToPanel(panel, _paramRegion.getEditor().getLabelComponent(), gbc,
        						"gridwidth=3, fill=BOTH, weightx=1");
        gbc.gridy = ++line;
        GridBagUtils.addToPanel(panel, _paramRegion.getEditor().getComponent(), gbc,
        						"insets.top=3, gridwidth=3, fill=BOTH, anchor=WEST");
 
        gbc.gridy = ++line;
        GridBagUtils.addToPanel(panel, _paramExpression.getEditor().getLabelComponent(), gbc,
        						"insets.top=4, gridwidth=3, anchor=WEST");
        gbc.gridy = ++line;
        GridBagUtils.addToPanel(panel, _paramExpression.getEditor().getComponent(), gbc,
        						"weighty=1, insets.top=3, gridwidth=3, fill=BOTH, anchor=WEST");
            
        gbc.gridy = ++line;
        GridBagUtils.addToPanel(panel, _paramUseExpression.getEditor().getComponent(), gbc,
        						"weighty=0, weightx=0, fill=NONE, gridwidth=1, anchor=EAST");
        GridBagUtils.addToPanel(panel, _editExpressionButton, gbc,
        						"insets.top=3, gridwidth=3, fill=NONE, anchor=EAST");    

        setContent(panel);
    }
	 
	   private void updateUIState() {
        final boolean useExpression = ((Boolean) _paramUseExpression.getValue()).booleanValue();
        _paramExpression.setUIEnabled(useExpression);
        _editExpressionButton.setEnabled(useExpression);
        
    }


	/** 
	 * Assignes the userinput to the com.bc.beam.visat.RegionAroundPin object and hides the dialog window, after the user has pressed the OK-button.
	 */
	protected void onOK(){
	 	
	 	final String[] extractedRegion =  _paramRegion.getValueAsText().split("x");
	 	final int region = new Integer(extractedRegion[0].trim()).intValue();
	   	final boolean useExpression = ((Boolean) _paramUseExpression.getValue()).booleanValue();
	 	String validExpression = null;
	 	
	 	if (useExpression) {
	 		validExpression = _paramExpression.getValueAsText();
	 	}
			try {
				exportRegion = new RegionAroundPin(_product,_selectedPin,region,validExpression);
				if (exportRegion.getNumberOfRelevantPixels()==0) {
					showErrorDialog("There are no pixels to export.");
					return;
				}
			} catch (IOException e) {
				showErrorDialog("An I/O error occured:\n" + e.getMessage());
				return;
			} catch (ParseException e) {
				showErrorDialog("Please check the expression you have entered.\nIt is not valid.");
				return;
			}

		super.onOK();
		
	 }
	/** 
	 * Gets called when the user has pressed the Cancel-button.
	 */
	protected void onCancel() {
        super.onCancel();
    }
	   
	/** 
	 * Checks the expression if the "Use Expression"-Parameter is set.
	 */
	protected boolean verifyUserInput() {       
	   	final boolean useExpression = ((Boolean) _paramUseExpression.getValue()).booleanValue();
   		if (!useExpression) {
   	        return super.verifyUserInput();
   		}
	   	if (!isValidExpression()) {
   			showErrorDialog("Please check the expression you have entered.\nIt is not valid."); 
   			return false;
   		}	   	
	    return super.verifyUserInput();

    }
	 
	     
	/**
	 * Computes the maximum size of the region.
	 * @param pin	the center of the region
	 * @param hMax	the product's height
	 * @param wMax	the product's width
	 * @return	the maximum height/width for the quadratic region around the pin
	 */
	private int computeMaxRegionSize(final Pin pin,final int hMax,final int wMax) {
    	final int pinPosX = (int) pin.getPixelPos().getX();
    	final int pinPosY = (int) pin.getPixelPos().getY();
    	final int xMin = Math.min(pinPosX,wMax-pinPosX);
    	final int yMin = Math.min(pinPosY,hMax-pinPosY);
    	final int n = Math.min(xMin,yMin);
    	
    	return n;
    }
	
    /**
     * Defines an array with all possible region sizes for the pulldown menu.
     * @param maxRegionSize	the maximum height/width for the quadratic region around the pin
     * @return the array with the possible options
     */
    private String[] defineRegionSelection(int maxRegionSize) {
		
    	// the maximal selectable size of the region is set to 25 x 25 pixels.
    	if (maxRegionSize > 25) {
			maxRegionSize = 25;
		}
    	final String[] regionSelection = new String[(maxRegionSize-1)/2];
		for (int i = 0;(3+i*2<=maxRegionSize);i++) {
			final int n = (3 + i*2);
    		regionSelection[i] = (n+" x "+n+" pixels" ); 
		}
    	return regionSelection;
    	
    }
 

    /**
     * Defines the ActionListener for the "Edit Expression"-button. 
     */
    private ActionListener createEditExpressionButtonListener() {
        return new ActionListener() {

            public void actionPerformed(final ActionEvent e) {
                final Product[] products = {_product};
            	ProductExpressionPane pep = ProductExpressionPane.createBooleanExpressionPane(products,_product,
                                                                                              _visatApp.getPreferences());
                pep.setCode(_paramExpression.getValueAsText());
                final int status = pep.showModalDialog(getJDialog(), "Arithmetic Expression Editor");
                if (status == ModalDialog.ID_OK) {
                    _paramExpression.setValue(pep.getCode(), null);
                    Debug.trace("BandArithmetikDialog: expression is: " + pep.getCode());
                }
                pep.dispose();
                pep = null;
            }
        };
    }

    private ParamChangeListener createParamChangeListener() {
        return new ParamChangeListener() {

            public void parameterValueChanged(final ParamChangeEvent event) {
                updateUIState();
            }
        };
    }
    
    /**
     * Checks the arithmetic expression.
     */
    private boolean isValidExpression() {
        final String expression = _paramExpression.getValueAsText();
        if (expression == null || expression.length() == 0) {
            return false;
        }
        final Parser parser = _product.createBandArithmeticParser();
        try {
            parser.parse(expression);
        } catch (ParseException e) {
            return false;
        }
        return true;
    }

}
