package com.bc.beam.visat;

import java.awt.Point;
import java.io.IOException;

import org.esa.beam.framework.datamodel.Pin;
import org.esa.beam.framework.datamodel.Product;
import org.esa.beam.framework.dataop.barithm.RasterDataEvalEnv;
import org.esa.beam.framework.dataop.barithm.RasterDataLoop;

import com.bc.jexp.ParseException;
import com.bc.jexp.Parser;
import com.bc.jexp.Term;

/*
 * Copyright (C) 2005  by Maximilian Aulinger
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/**
 * Represents a region of <code>size x size</code> pixels around a Pin.
 * An arithmetic expression can be used to filter the pixels.
 * 
 * @author Maximilian Aulinger
 *
 */

public class RegionAroundPin {
	private Product _product;
	private Pin centerOfRegion;
	private Point firstPixel;
	private Point lastPixel;
	private int size;
	private String expression;
	private boolean[] relevantPixels;
	
	
	/**
	 * Initialises a new instance of this class.
	 * @param product			the product providing the pixel values
	 * @param pin				a pin defining the center of the region
	 * @param regionSize		the height/width of the quadratic region
	 * @param validExpression	an arithmetic expression
	 * @throws IOException		if an error occures while reading the product's datalines
	 * @throws ParseException	if the given expression is invalid
	 */
	public RegionAroundPin(final Product product,final Pin pin,final int regionSize,final String validExpression) throws IOException, ParseException {
		_product = product;
		centerOfRegion = pin;
		size = regionSize;
		expression = validExpression;
		firstPixel = computeFirstPixel(centerOfRegion,size);
		lastPixel = computeLastPixel(centerOfRegion,size);
		if (expression != null) {
			relevantPixels = computeRelevantPixels(_product,expression);
		}
	}

	/**
	 * Returns the userdefinded expression, <code>null</code> if no expression is defined.  
	 */
	public String getExpression() {
		return expression;
	}
	/**
	 * Returns the center of the region. The center of the region is defined by a single pin.
	 */
	public Pin getCenterOfRegion() {
		return centerOfRegion;
	}
	/**
	 * Returns the first pixel in the upper left corner of the region.
	 */
	public Point getFirstPixel() {
		return firstPixel;
	}
	/**
	 * Returns the last pixel in the lower right corner of the region.
	 */
	public Point getLastPixel() {
		return lastPixel;
	}
	/**
	 * Returns the size of the region in pixels.
	 */
	public int getSize() {
		return size;
	}
	
	/**
	 * Returns the pixel size of the region as a String.
	 */
	public String getRegion() {
		return size+" x "+size;
	}
	
	/**
	 * Returns the number of relevant pixels.
	 */
	public int getNumberOfRelevantPixels() {
		int numRelPix = 0;
		for (int i = (int) firstPixel.getY();i<=lastPixel.getY();i++) {
			for (int j = (int) firstPixel.getX();j<=lastPixel.getX();j++) {
				if (isRelevantPixel(j,i)) {
					numRelPix++;
				}
			}
		}
		
		return numRelPix;
	}

	/**
	 * Checks whether the given pixel is part of the region an is not filtered by
	 * an arithmetic expression.
	 * @param x the x coordinate of the pixel 
	 * @param y the y coordinate of the pixel
	 * @return <code>true</code> if the given pixel is included
	 * in the region and is not filtered by an arithmetic expression 
	 */
	public boolean isRelevantPixel(final int x,final int y) {
		final boolean pixelIsInRegion = contains(x,y);
		if (expression != null) {
			if (pixelIsInRegion) {
				final int pixelIndex = (x - firstPixel.x) + (y - firstPixel.y)*size;
				return relevantPixels[pixelIndex];
			}
		}
		return pixelIsInRegion;
	}

	/**
	 * Checks whether the given pixel is part of the region.
	 * @param x the x coordinate of the pixel
	 * @param y the y coordinate of the pixel
	 * @return <code>true</code> if the given pixel is included in the region
	 */
	public boolean contains(final int x,final int y){
		if((x >= (int) firstPixel.getX()) && (x <= (int)lastPixel.getX())) { 
				if ((y >= (int)firstPixel.getY()) && (y <= (int)lastPixel.getY())) {
					return true;
				}
		}
			return false;
	}

	/**
	 * Computes the position of the first pixel in the upper left corner of the region.
	 */
	private Point computeFirstPixel(final Pin center,final int size) {
		int x = (int) center.getPixelPos().getX();
		int y = (int) center.getPixelPos().getY();
		x = x - (size-1)/2;
		y = y - (size-1)/2;
		final Point firstPixel = new Point(x,y);
		return firstPixel;
	}
	/**
	 * Computes the position of the last pixel in the lower right corner of the region.
	 */
	private Point computeLastPixel(final Pin center,final int size) {
		int x = (int) center.getPixelPos().getX();
		int y = (int) center.getPixelPos().getY();
		x = x + (size-1)/2;
		y = y + (size-1)/2;
		final Point lastPixel = new Point(x,y);
		return lastPixel;		
		}
	
	/**
	 * Evaluates the arithmetic expression for every pixel in the given region.
	 */
	private boolean[] computeRelevantPixels(final Product selectedProduct,final String expression) throws IOException, ParseException  {
    	final Parser parser = selectedProduct.createBandArithmeticParser();
        final Term term;
		term = parser.parse(expression);

		final int regionWidth = lastPixel.x - firstPixel.x + 1;
		final int regionHeight = lastPixel.y - firstPixel.y + 1;
		final boolean[] relevantPixels = new boolean[regionWidth * regionHeight];
		final RasterDataEvalEnv env = new RasterDataEvalEnv(firstPixel.x, firstPixel.y, regionWidth, regionHeight);
		final RasterDataLoop loop = new RasterDataLoop(env,new Term[]{term});
		       
		loop.forEachPixel(new RasterDataLoop.Body(){
		    public void eval(final RasterDataEvalEnv env, final int i) {
		         relevantPixels[i] = term.evalB(env);
		    }
		});
		
		return relevantPixels;
		
	}
	
}
